# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Python language schemas utilities."""

from re import compile as re_compile

from b3j0f.utils.version import OrderedDict
from b3j0f.utils.path import lookup

from ..base import Schema, DynamicValue
from .factory import SchemaBuilder, build
from ..elementary import ElementarySchema, ArraySchema, OneOfSchema, TypeSchema
from ..utils import updatecontent, data2schema, datatype2schemacls, RefSchema

from types import (
    FunctionType, MethodType, LambdaType, BuiltinFunctionType,
    BuiltinMethodType, MemberDescriptorType
)

from six import get_function_globals

from inspect import getargspec, getsourcelines, isclass, isbuiltin

from functools import wraps

__all__ = [
    'PythonSchemaBuilder', 'FunctionSchema', 'buildschema', 'ParamSchema'
]


class PythonSchemaBuilder(SchemaBuilder):
    """In charge of building python classes."""

    __name__ = 'python'

    def build(self, _resource, **kwargs):

        if not isclass(_resource):
            raise TypeError(
                'Wrong type {0}, \'type\' expected'.format(_resource)
            )

        if issubclass(_resource, Schema):
            result = _resource

        else:
            result = datatype2schemacls(_datatype=_resource, _force=False)

            if result is None:
                resname = _resource.__name__

                if 'name' not in kwargs:
                    kwargs['name'] = resname

                for attrname in dir(_resource):
                    if (
                        attrname and attrname[0] != '_' and
                        attrname not in kwargs and
                        not hasattr(Schema, attrname)
                    ):
                        attr = getattr(_resource, attrname)

                        if not isinstance(attr, MemberDescriptorType):
                            kwargs[attrname] = attr

                result = type(resname, (Schema,), kwargs)

        return result

    def getresource(self, schemacls):

        result = None

        if hasattr(schemacls, 'mro'):
            for mro in schemacls.mro():

                if issubclass(mro, Schema):
                    result = mro
                    break

        return result


def buildschema(_cls=None, **kwargs):
    """Class decorator used to build a schema from the decorate class.

    :param type _cls: class to decorate.
    :param kwargs: schema attributes to set.
    :rtype: type
    :return: schema class.
    """
    if _cls is None:
        return lambda _cls: buildschema(_cls=_cls, **kwargs)

    result = build(_cls, **kwargs)

    return result


class ParamTypeSchema(Schema):
    """In charge of embedding a parameter type which met a problem while
    generating a schema."""

    type = TypeSchema()

    def _validate(self, data, *args, **kwargs):

        super(ParamTypeSchema, self)._validate(data=data, *args, **kwargs)

        if not isinstance(data, self.type):
            raise TypeError(
                'Wrong type of {0}. {1} expected.'.format(data, self.type)
            )


@updatecontent
class ParamSchema(RefSchema):
    """Function parameter schema."""

    #: if true (default), update self ref when default is given.
    autotype = True
    mandatory = True  #: if true (default), parameter value is mandatory.

    def _setvalue(self, schema, value, *args, **kwargs):

        super(ParamSchema, self)._setvalue(schema, value, *args, **kwargs)

        if schema.name == 'default':

            if self.autotype and self.ref is None:
                self.ref = None if value is None else data2schema(value)

            if value is not None:
                self.mandatory = False


class FunctionSchema(ElementarySchema):
    """Function schema.

    Dedicated to describe functions, methods and lambda objects.
    """

    _PDESC = r':param (?P<ptype1>[\w_,]+) (?P<pname1>\w+):'
    _PTYPE = r':type (?P<pname2>[\w_]+):(?P<ptype2>[^\n]+)'
    _RTYPE = r':rtype:(?P<rtype>[^\n]+)'

    _REC = re_compile('{0}|{1}|{2}'.format(_PDESC, _PTYPE, _RTYPE))

    __data_types__ = [
        FunctionType, MethodType, LambdaType, BuiltinFunctionType,
        BuiltinMethodType
    ]

    params = ArraySchema(itemtype=ParamSchema())
    rtype = Schema()
    impl = ''
    impltype = ''
    safe = False
    varargs = False

    def _validate(self, data, owner, *args, **kwargs):

        ElementarySchema._validate(self, data=data, *args, **kwargs)

        if data != self.default or data is not self.default:

            errormsg = 'Error while validating {0} with {1}'.format(data, self)

            if data.__name__ != self.name:

                raise TypeError(
                    '{0}. Wrong function name {1}. {2} expected.'.format(
                        errormsg, data.__name__, self.name
                    )
                )

            params, rtype, vargs, kwargs = self._getparams_rtype(function=data)

            var = self.varargs or vargs or kwargs

            if (not var) and len(params) != len(self.params):
                raise TypeError(
                    '{0}. Wrong param length: {1}. {2} expected.'.format(
                        errormsg, len(params), len(self.params)
                    )
                )

            if self.rtype is not None and type(self.rtype) != type(rtype):
                raise TypeError(
                    '{0}. Wrong rtype {1}. {2} expected.'.format(
                        rtype, self.rtype
                    )
                )

            for index, pkwargs in enumerate(params.values()):
                name = pkwargs['name']
                default = pkwargs.get('default')
                param = self.params[index]

                if param.name != name:
                    raise TypeError(
                        '{0}. Wrong param {1} at {2}. {3} expected.'.format(
                            errormsg, name, index, param.name
                        )
                    )

                val = param.default
                if isinstance(val, DynamicValue):
                    val = val()

                if (
                    val is not None and default is not None and val != default
                ):
                    raise TypeError(
                        '{0}. Wrong val {1}/{2} at {3}. Expected {4}.'.format(
                            errormsg, name, default, index, val
                        )
                    )

    def _setvalue(self, schema, value):

        if schema.name == 'default':
            self._setter(obj=self, value=value)

    def _setter(self, obj, value, *args, **kwargs):

        if hasattr(self, 'olddefault'):
            if self.olddefault is value:
                return

        self.olddefault = value

        ElementarySchema._setter(self, obj, value, *args, **kwargs)

        pkwargs, self.rtype, vargs, kwargs = self._getparams_rtype(value)

        self.vargs = vargs or kwargs

        params = []

        selfparams = {}

        for selfparam in self.params:
            selfparams[selfparam.name] = selfparam

        index = 0

        for index, pkwarg in enumerate(pkwargs.values()):

            name = pkwarg['name']

            selfparam = None  # old self param

            if name in selfparams:
                selfparam = selfparams[name]

            if selfparam is None:
                selfparam = ParamSchema(**pkwarg)

            else:
                for key in pkwarg:
                    val = pkwarg[key]
                    if val is not None:
                        setattr(selfparam, key, val)

            params.append(selfparam)

        self.params = params

        self.impltype = 'python'

        try:
            self.impl = str(getsourcelines(value))

        except TypeError:
            self.impl = ''

    @classmethod
    def _getparams_rtype(cls, function):
        """Get function params from input function and rtype.

        :return: OrderedDict, rtype, vargs and kwargs.
        :rtype: tuple
        """
        try:
            args, vargs, kwargs, default = getargspec(function)

        except TypeError:
            args, vargs, kwargs, default = (), (), (), ()

        indexlen = len(args) - (0 if default is None else len(default))

        params = OrderedDict()

        for index, arg in enumerate(args):

            pkwargs = {
                'name': arg,
                'mandatory': True
            }  # param kwargs

            if index >= indexlen:  # has default value
                value = default[index - indexlen]
                pkwargs['default'] = value
                pkwargs['ref'] = None if value is None else data2schema(value)
                pkwargs['mandatory'] = False

            params[arg] = pkwargs

        rtype = None

        # parse docstring
        if function.__doc__ is not None and not isbuiltin(function):

            scope = get_function_globals(function)

            for match in cls._REC.findall(function.__doc__):
                if rtype is None:
                    rrtype = match[4].strip() or None

                    if rrtype:

                        rtypes = rrtype.split(',')

                        schemas = []
                        for rtype_ in rtypes:

                            rtype_ = rtype_.strip()
                            islist = False

                            try:
                                lkrtype = lookup(rtype_, scope=scope)

                            except ImportError:
                                islist = True

                                try:
                                    if rtype_[-1] == 's':
                                        lkrtype = lookup(
                                            rtype_[:-1], scope=scope
                                        )

                                    elif rtype_.startswith('list of '):
                                        lkrtype = lookup(
                                            rtype_[8:], scope=scope
                                        )

                                    else:
                                        raise

                                except ImportError:
                                    msg = 'rtype "{0}" ({1}) from {2} not found.'
                                    raise ImportError(
                                        msg.format(rtype_, rrtype, function)
                                    )

                            try:
                                schemacls = datatype2schemacls(lkrtype)

                            except TypeError:
                                schemacls = ParamTypeSchema(type=lkrtype)

                            rschema = schemacls()

                            if islist:
                                rschema = ArraySchema(itemtype=rschema)

                            schemas.append(rschema)

                        if len(rtypes) > 1:
                            rtype = OneOfSchema(schemas=schemas, nullable=True)

                        else:
                            rtype = schemas[0]

                        continue

                pname = (match[1] or match[2]).strip()

                if pname and pname in params:

                    ptype = (match[0] or match[3]).strip()

                    ptypes = ptype.split(',')

                    schemas = []

                    for ptype in ptypes:

                        ptype = ptype.strip()

                        islist = False

                        try:
                            lkptype = lookup(ptype, scope=scope)

                        except ImportError:
                            islist = True

                            try:
                                if ptype[-1] == 's':
                                    lkptype = lookup(ptype[:-1], scope=scope)

                                elif ptype.startswith('list of '):
                                    lkptype = lookup(ptype[8:], scope=scope)

                                else:
                                    raise

                            except ImportError:

                                msg = 'Error on ptype "{0}" ({1}) from {2} not found.'
                                raise ImportError(
                                    msg.format(pname, ptype, function)
                                )

                        try:
                            schemacls = datatype2schemacls(lkptype)

                        except TypeError:
                            schemacls = ParamTypeSchema(type=lkptype)

                        pschema = schemacls()

                        if islist:
                            pschema = ArraySchema(itemtype=pschema)

                        schemas.append(pschema)

                    if len(ptypes) > 1:
                        pschema = OneOfSchema(schemas=schemas, nullable=True)

                    else:
                        pschema = schemas[0]

                    params[pname]['ref'] = pschema

        return params, rtype, vargs, kwargs

    def __call__(self, *args, **kwargs):

        return self.default(*args, **kwargs)

    def _getter(self, obj, *args, **kwargs):

        func = ElementarySchema._getter(self, obj, *args, **kwargs)

        @wraps(func)
        def result(*args, **kwargs):

            try:
                result = func(obj, *args, **kwargs)

            except TypeError:
                result = func(*args, **kwargs)

            return result

        result.source = func

        return result


def funcschema(default=None, *args, **kwargs):
    """Decorator to use in order to transform a function into a schema."""
    if default is None:
        return lambda default: funcschema(default=default, *args, **kwargs)

    return FunctionSchema(default=default, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Factory UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..factory import SchemaFactory, SchemaBuilder, getbuilder


class TestFactory(UTCase):

    def setUp(self):

        self.factory = SchemaFactory()

        self.builder = lambda _type: TestFactory.SchemaBuilderTest(_type)

    class SchemaBuilderTest(SchemaBuilder):

        __register__ = False

        def __init__(self, _type, *args, **kwargs):

            super(TestFactory.SchemaBuilderTest, self).__init__(
                *args, **kwargs
            )

            self.type = _type

        def build(self, _resource, **kwargs):

            if not isinstance(_resource, self.type):
                raise TypeError()

            return _resource

        def getresource(self, schemacls):

            return schemacls

    def test_register(self):

        makerstr = self.builder(str)
        makerint = self.builder(int)

        schemastr = 'test'
        schemaint = 2

        self.factory = SchemaFactory()

        self.assertRaises(ValueError, self.factory.build, schemaint)
        self.assertIsNone(self.factory.getschemacls(schemaint))
        self.assertRaises(ValueError, self.factory.build, schemastr)
        self.assertIsNone(self.factory.getschemacls(schemastr))

        self.factory.registerbuilder(name='str', builder=makerstr)

        self.assertRaises(ValueError, self.factory.build, schemaint)
        self.assertIsNone(self.factory.getschemacls(schemaint))

        schemacls = self.factory.build(schemastr)
        self.assertEqual(schemacls, schemastr)

        schemacls = self.factory.getschemacls(schemastr)
        self.assertEqual(schemacls, schemastr)

        self.factory.registerbuilder(name='int', builder=makerint)

        schemacls = self.factory.build(schemaint)
        self.assertEqual(schemacls, schemaint)

        schemacls = self.factory.getschemacls(schemaint)
        self.assertEqual(schemacls, schemaint)

        schemacls = self.factory.build(schemastr)
        self.assertEqual(schemacls, schemastr)

        schemacls = self.factory.getschemacls(schemastr)
        self.assertEqual(schemacls, schemastr)

        self.factory.unregisterbuilder(name='str')

        schemacls = self.factory.build(schemastr)
        self.assertEqual(schemacls, schemastr)

        self.assertRaises(
            ValueError, self.factory.build, schemastr, _cache=False
        )

        schemacls = self.factory.getschemacls(schemastr)
        self.assertEqual(schemacls, schemastr)

        schemacls = self.factory.build(schemaint)
        self.assertEqual(schemacls, schemaint)

        schemacls = self.factory.build(schemastr)
        self.assertEqual(schemacls, schemastr)

    def test_autoregister(self):

        class TestSchemaBuilder(SchemaBuilder):
            pass

        builder = getbuilder(TestSchemaBuilder.__name__)

        self.assertIsInstance(builder, TestSchemaBuilder)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Schema utilities package."""

from six import iteritems, add_metaclass

from inspect import getmembers

from .registry import getbydatatype, register
from .lang.factory import build, getschemacls
from .base import Schema, DynamicValue

__all__ = [
    'DynamicValue', 'data2schema', 'MetaRegisteredSchema', 'ThisSchema',
    'updatecontent', 'validate', 'dump', 'RegisteredSchema',
    'datatype2schemacls', 'RefSchema', 'data2schemacls', 'AnySchema'
]


class AnySchema(Schema):
    """Schema for any data."""

    def _validate(*args, **kwargs):
        pass


def datatype2schemacls(
        _datatype, _registry=None, _factory=None, _force=True,
        _besteffort=True, **kwargs
):
    """Get a schema class which has been associated to input data type by the
    registry or the factory in this order.

    :param type datatype: data type from where get associated schema.
    :param SchemaRegisgry _registry: registry from where call the getbydatatype
        . Default is the global registry.
    :param SchemaFactory _factory: factory from where call the getschemacls if
        getbydatatype returns None. Default is the global factory.
    :param bool _force: if true (default), force the building of schema class
        if no schema is associated to input data type.
    :param bool _besteffort: if True (default), try to resolve schema by
        inheritance.
    :param dict kwargs: factory builder kwargs.
    :rtype: type
    :return: Schema associated to input registry or factory. None if no
        association found.
    """
    result = None

    gdbt = getbydatatype if _registry is None else _registry.getbydatatype

    result = gdbt(_datatype, besteffort=_besteffort)

    if result is None:
        gscls = getschemacls if _factory is None else _factory.getschemacls
        result = gscls(_datatype, besteffort=_besteffort)

    if result is None and _force:
        _build = build if _factory is None else _factory.build

        result = _build(_resource=_datatype, **kwargs)

    return result


def data2schema(
        _data=None, _force=False, _besteffort=True, _registry=None,
        _factory=None, _buildkwargs=None, **kwargs
):
    """Get the schema able to instanciate input data.

    The default value of schema will be data.

    Can be used such as a decorator:

    ..code-block:: python

        @data2schema
        def example(): pass  # return a function schema

        @data2schema(_registry=myregistry)
        def example(): pass  # return a function schema with specific registry

    ..warning::

        return this function id _data is None.

    :param _data: data possibly generated by a schema. Required but in case of
        decorator.
    :param bool _force: if True (False by default), create the data schema
        on the fly if it does not exist.
    :param bool _besteffort: if True (default), find a schema class able to
        validate data class by inheritance.
    :param SchemaRegistry _registry: default registry to use. Global by
        default.
    :param SchemaFactory factory: default factory to use. Global by default.
    :param dict _buildkwargs: factory builder kwargs.
    :param kwargs: schema class kwargs.
    :return: Schema.
    :rtype: Schema.
    """
    if _data is None:
        return lambda _data: data2schema(
            _data, _force=False, _besteffort=True, _registry=None,
            _factory=None, _buildkwargs=None, **kwargs
        )

    result = None

    fdata = _data() if isinstance(_data, DynamicValue) else _data

    datatype = type(fdata)

    content = getattr(fdata, '__dict__', {})
    if _buildkwargs:
        content.udpate(_buildkwargs)

    schemacls = datatype2schemacls(
        _datatype=datatype, _registry=_registry, _factory=_factory,
        _force=_force, _besteffort=_besteffort, **content
    )

    if schemacls is not None:
        result = schemacls(default=_data, **kwargs)

        for attrname in dir(_data):

            if not hasattr(schemacls, attrname):
                attr = getattr(_data, attrname)
                if attr is not None:
                    setattr(result, attrname, attr)

    if result is None and _data is None:
        result = AnySchema()

    return result


def data2schemacls(_data, **kwargs):
    """Convert a data to a schema cls.

    :param data: object or dictionary from where get a schema cls.
    :return: schema class.
    :rtype: type
    """
    content = {}

    for key in list(kwargs):  # fill kwargs
        kwargs[key] = data2schema(kwargs[key])

    if isinstance(_data, dict):
        datacontent = iteritems(_data)

    else:
        datacontent = getmembers(_data)

    for name, value in datacontent:

        if name[0] == '_':
            continue

        if isinstance(value, dict):
            schema = data2schemacls(value)()

        else:
            schema = data2schema(value)

        content[name] = schema

    content.update(kwargs)  # update content

    result = type('GeneratedSchema', (Schema,), content)

    return result


class ThisSchema(object):
    """Tool Used to set inner schemas with the same type with specific arguments
    .

    ThisSchema one might be use at the condition instanciation methods must not
    reference the class.

    ..example::

        class Test(Schema):
            # contain an inner schema nullable 'test' of type Test.
            test = ThisSchema(nullable=False)

            def __init__(self, *args, **kwargs):

                # old style call because when the schema will be automatically
                # updated, the class Test does not exist in the scope
                Schema.__init__(self, *args, **kwargs)

    :param args: schema class vargs to use.
    :param kwargs: schema class kwargs to use.

    :return: input args and kwargs.
    :rtype: tuple
    """

    def __init__(self, *args, **kwargs):

        super(ThisSchema, self).__init__()

        self.args = args
        self.kwargs = kwargs


def validate(schema, data, owner=None):
    """Validate input data with input schema.

    :param Schema schema: schema able to validate input data.
    :param data: data to validate.
    :param Schema owner: input schema parent schema.
    :raises: Exception if the data is not validated.
    """
    schema._validate(data=data, owner=owner)


def dump(schema):
    """Get a serialized value of input schema.

    :param Schema schema: schema to serialize.
    :rtype: dict
    """
    result = {}

    for name, _ in iteritems(schema.getschemas()):

        if hasattr(schema, name):
            val = getattr(schema, name)

            if isinstance(val, DynamicValue):
                val = val()

            if isinstance(val, Schema):
                val = dump(val)

            result[name] = val

    return result


class RefSchema(Schema):
    """Schema which references another schema."""

    ref = Schema(name='ref')  #: the reference must be a schema.

    def __init__(self, ref=None, *args, **kwargs):
        """
        :param Schema ref: refereed schema.
        """

        super(RefSchema, self).__init__(*args, **kwargs)

        if ref is not None:

            if self.default is None:
                self.default = ref.default

            self.ref = ref

    def _validate(self, data, owner=None, *args, **kwargs):

        ref = owner if self.ref is None else self.ref

        if ref is not self:
            ref._validate(data=data, owner=owner, *args, **kwargs)

    def _setvalue(self, schema, value, *args, **kwargs):

        super(RefSchema, self)._setvalue(schema, value, *args, **kwargs)

        if schema.name == 'ref' and value is not None:

            if self.default is None and not value.nullable:

                self.default = value.default

            value._validate(self.default)


def updatecontent(schemacls=None, updateparents=True, exclude=None):
    """Transform all schema class attributes to schemas.

    It can be used such as a decorator in order to ensure to update attributes
    with the decorated schema but take care to the limitation to use old style
    method call for overidden methods.

    .. example:
        @updatecontent  # update content at the end of its definition.
        class Test(Schema):
            this = ThisSchema()  # instance of Test.
            def __init__(self, *args, **kwargs):
                Test.__init__(self, *args, **kwargs)  # old style method call.

    :param type schemacls: sub class of Schema.
    :param bool updateparents: if True (default), update parent content.
    :param list exclude: attribute names to exclude from updating.
    :return: schemacls.
    """
    if schemacls is None:
        return lambda schemacls: updatecontent(
            schemacls=schemacls, updateparents=updateparents, exclude=exclude
        )

    if updateparents and hasattr(schemacls, 'mro'):
        schemaclasses = reversed(list(schemacls.mro()))

    else:
        schemaclasses = [schemacls]

    for schemaclass in schemaclasses:
        for name, member in iteritems(getattr(schemaclass, '__dict__', {})):
            # transform only public members
            if name[0] != '_' and (exclude is None or name not in exclude):

                toset = False  # flag for setting schemas

                fmember = member

                if isinstance(fmember, DynamicValue):
                    fmember = fmember()
                    toset = True

                if isinstance(fmember, Schema):
                    schema = fmember

                    if not schema.name:
                        schema.name = name

                else:
                    toset = True

                    data = member

                    if name == 'default':

                        if isinstance(fmember, ThisSchema):
                            data = schemaclass(*fmember.args, **fmember.kwargs)

                        schema = RefSchema(default=data, name=name)

                    elif isinstance(fmember, ThisSchema):

                        schema = schemaclass(
                            name=name, *fmember.args, **fmember.kwargs
                        )

                    elif member is None:
                        schema = AnySchema(name=name)

                    else:
                        schema = data2schema(_data=data, name=name)

                if isinstance(schema, Schema) and toset:

                    try:
                        setattr(schemaclass, name, schema)

                    except (AttributeError, TypeError):
                        break

    return schemacls

updatecontent(RefSchema)


class MetaRegisteredSchema(type):
    """Automatically register schemas."""

    def __new__(mcs, *args, **kwargs):

        result = super(MetaRegisteredSchema, mcs).__new__(mcs, *args, **kwargs)

        # update all sub schemas related to values
        if result.__update_content__:
            updatecontent(schemacls=result)

        return result

    def __call__(cls, *args, **kwargs):

        result = super(MetaRegisteredSchema, cls).__call__(*args, **kwargs)

        if result.__register__:  # register all new schema
            register(schema=result)

        return result


# use metaRegisteredschema such as the schema metaclass
@add_metaclass(MetaRegisteredSchema)
class RegisteredSchema(Schema):
    """Ease auto-registering of schemas and auto-updating content."""

    #: Register instances in the registry if True (False by default).
    __register__ = False

    """update automatically the content if True (default).

    If True, take care to not having called the class in overidden methods.
    In such case, take a look to the using of the class ThisSchema which
    recommands to use old style method call for overriden methods.

    ..example:
        class Test(Schema):
            __udpate_content__ = True  # set update content to True
            test = ThisSchema()
            def __init__(self, *args, **kwargs):
                Schema.__init__(self, *args, **kwargs)  # old style call.
        """
    __update_content__ = True



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Schema factory module."""

from uuid import uuid4 as uuid

from six import add_metaclass

from inspect import isclass

__all__ = [
    'SchemaFactory', 'registerbuilder', 'unregisterbuilder', 'build',
    'getbuilder', 'getschemacls', 'getresource', 'SchemaBuilder'
]


class SchemaFactory(object):
    """Factory dedicated to generate schemas.

    Keys are schema builder names, and values are schema builders.
    """

    def __init__(self, builders=None, schemasbyresource=None, *args, **kwargs):

        super(SchemaFactory, self).__init__(*args, **kwargs)

        self._schemasbyresource = schemasbyresource or {}
        self._builders = builders or {}

    def registerbuilder(self, builder, name=None):
        """Register a schema builder with a key name.

        Can be used such as a decorator where the builder can be the name for a
        short use.

        :param SchemaBuilder builder: schema builder.
        :param str name: builder name. Default is builder name or generated.
        """
        if name is None:
            name = uuid()

        self._builders[name] = builder

        return builder

    def unregisterbuilder(self, name):
        """Unregister a builder by its name.

        :param str name: builder name to remove.
        :raises: KeyError if name is not registered.
        """
        del self._builders[name]

    @property
    def builders(self):
        """Get builder names.

        :rtype: list
        """
        return list(self._builders.keys())

    def getbuilder(self, name):
        """Get a builder instance from a name.

        :param str name: builder name to retrieve.
        """
        return self._builders[name]

    def build(self, _resource, _cache=True, updatecontent=True, **kwargs):
        """Build a schema class from input _resource.

        :param _resource: object from where get the right schema.
        :param bool _cache: use _cache system.
        :param bool updatecontent: if True (default) update result.
        :rtype: Schema.
        """
        result = None

        if _cache and _resource in self._schemasbyresource:
            result = self._schemasbyresource[_resource]

        else:
            for builder in self._builders.values():
                try:
                    result = builder.build(_resource=_resource, **kwargs)

                except Exception:
                    pass

                else:
                    break

            if result is None:
                raise ValueError('No builder found for {0}'.format(_resource))

            if _cache:
                self._schemasbyresource[_resource] = result

            if updatecontent:
                from ..utils import updatecontent
                updatecontent(result, updateparents=False)

        return result

    def getschemacls(self, resource, besteffort=True):
        """Get schema class related to input resource.

        :param resource: resource from which get schema class.
        :param bool besteffort: if True (default) try a best effort in parsing
            the inheritance tree of resource if resource is a class.
        :rtype: type
        """
        result = None

        if hasattr(resource, 'mro') and besteffort and isclass(resource):
            resources = list(resource.mro())

        else:
            resources = [resource]

        for _resource in resources:
            if _resource in self._schemasbyresource:
                result = self._schemasbyresource[_resource]

                break

        return result

    def getresource(self, schemacls, name):
        """Get a resource from a builder name.

        :param type schemacls: waited schema class.
        :param str name: builder name to use.
        :return: resource returned by the right builder.getresource(schema).
        """
        return self.builders[name].getresource(schemacls=schemacls)

_SCHEMAFACTORY = SchemaFactory()  #: global schema factory


def registerbuilder(builder, name=None):
    """Register a schema builder.

    :param str name: builder name. Default is builder name or generated.
    :param builder: callable object which takes in parameter a schema resource
        and generate a schema class in return. If the resource is not in the
        right format, the builder must raise a TypeError exception.
    """
    return _SCHEMAFACTORY.registerbuilder(name=name, builder=builder)


def unregisterbuilder(name):
    """Unregister a builder by its name.

    :param str name: builder name to remove.
    :raises: KeyError if name is not registered.
    """
    return _SCHEMAFACTORY.unregisterbuilder(name=name)


def build(_resource, _cache=True, **kwargs):
    """Build a schema from input _resource.

    :param _resource: object from where get the right schema.
    :param bool _cache: use cache system.
    :rtype: Schema.
    """
    return _SCHEMAFACTORY.build(_resource=_resource, _cache=True, **kwargs)


def getbuilder(name):
    """Get a builder instance from a name.

    :param str name: builder name to retrieve.
    """
    return _SCHEMAFACTORY.getbuilder(name)


def getschemacls(resource, besteffort=False):
    """Get schema class related to input resource.

    :param resource: resource from which get schema class.
    :param bool besteffort: if True (default) try a best effort in parsing
            the inheritance tree of resource if resource is a class.
    :rtype: type
    """
    return _SCHEMAFACTORY.getschemacls(
        resource=resource, besteffort=besteffort
    )


def getresource(self, schemacls, name):
    """Get a resource from a builder name.

    :param type schemacls: waited schema class.
    :param str name: builder name to use.
    :return: resource returned by the right builder.getresource(schema).
    """
    return _SCHEMAFACTORY.getresource(schemacls=schemacls, name=name)


class MetaSchemaBuilder(type):

    def __new__(cls, *args, **kwargs):

        result = super(MetaSchemaBuilder, cls).__new__(cls, *args, **kwargs)

        if result.__register__:
            registerbuilder(result(), name=result.__name__ or result.__name__)

        return result


@add_metaclass(MetaSchemaBuilder)
class SchemaBuilder(object):
    """Schema builder interface for building schema from a resource, and
    reciprocally.
    """

    __register__ = True  #: if True (default), automatically register this.
    __name__ = None  #: schema builder name. Default is generated.

    def build(self, _resource, **kwargs):
        """Build a schema class from input resource."""
        raise NotImplementedError()

    def getresource(self, schemacls):
        """Get a schema resource from input schema."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Main package."""

from .version import __version__

from .lang import (
    build, getresource, SchemaFactory, SchemaBuilder, getschemacls,
    XSDSchemaBuilder, FunctionSchema, ParamSchema
)
from .base import Schema, DynamicValue
from .registry import register, registercls, getbyuuid, getbyname
from .utils import (
    validate, dump, updatecontent, ThisSchema, data2schema, AnySchema,
    data2schemacls, RegisteredSchema, RefSchema
)
from .elementary import (
    StringSchema, DateTimeSchema, IntegerSchema, NumberSchema, LongSchema,
    ComplexSchema, FloatSchema, ArraySchema, DictSchema, TypeSchema,
    EnumSchema, BooleanSchema, ElementarySchema, OneOfSchema
)

__all__ = [
    '__version__', 'Schema', 'DynamicValue',
    'register', 'registercls', 'getbyuuid', 'getbyname',
    'StringSchema', 'IntegerSchema',
    'FloatSchema', 'ComplexSchema', 'EnumSchema', 'ArraySchema',
    'DictSchema', 'DateTimeSchema', 'NumberSchema', 'BooleanSchema',
    'TypeSchema', 'AnySchema', 'OneOfSchema',
    'data2schemacls',
    'data2schema', 'validate', 'dump', 'updatecontent', 'ThisSchema',
    'RefSchema', 'AnySchema', 'RegisteredSchema',
    'ElementarySchema', 'LongSchema',
    'SchemaFactory', 'getschemacls', 'getresource', 'SchemaBuilder', 'build',
    'XSDSchemaBuilder', 'FunctionSchema', 'ParamSchema'
]



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Elementary schema package."""

from six import string_types, add_metaclass, PY3

from numbers import Number

from enum import Enum

from datetime import datetime

from .base import Schema
from .registry import registercls
from .utils import (
    DynamicValue, RegisteredSchema, ThisSchema, MetaRegisteredSchema, validate,
    AnySchema
)

__all__ = [
    'NumberSchema',
    'IntegerSchema', 'FloatSchema', 'ComplexSchema', 'LongSchema',
    'StringSchema',
    'ArraySchema',
    'DictSchema',
    'BooleanSchema',
    'EnumSchema',
    'DateTimeSchema',
    'OneOfSchema'
]

NoneType = type(None)


class MetaElementarySchema(MetaRegisteredSchema):
    """Automatically register schemas with data types."""

    def __new__(mcs, *args, **kwargs):

        result = super(MetaElementarySchema, mcs).__new__(mcs, *args, **kwargs)

        if result.__data_types__:
            registercls(data_types=result.__data_types__, schemacls=result)

        return result


@add_metaclass(MetaElementarySchema)
class ElementarySchema(RegisteredSchema):
    """Base elementary schema."""

    __register__ = False

    nullable = False

    #: data types which can be instanciated by this schema.
    __data_types__ = []

    def _validate(self, data, owner=None, *args, **kwargs):
        """Validate input data in returning an empty list if true.

        :param data: data to validate with this schema.
        :param Schema owner: schema owner.
        :raises: Exception if the data is not validated.
        """
        if isinstance(data, DynamicValue):
            data = data()

        if data is None and not self.nullable:
            raise TypeError('Value can not be null')

        elif data is not None:
            # data must inherits from this data_types
            if not isinstance(data, tuple(self.__data_types__)):
                raise TypeError(
                    'Wrong data value: {0}. {1} expected.'.format(
                        data, self.__data_types__
                    )
                )


class BooleanSchema(ElementarySchema):
    """Boolean schema."""

    __data_types__ = [bool]
    default = False


class NumberSchema(ElementarySchema):
    """Schema for number such as float, long, complex and float.

    If allows to bound data values.
    """

    __data_types__ = [Number]
    default = 0

    #: minimum allowed value if not None.
    min = ThisSchema(nullable=True, default=None)
    #: maximal allowed value if not None.
    max = ThisSchema(nullable=True, default=None)

    def _validate(self, data, *args, **kwargs):

        ElementarySchema._validate(self, data, *args, **kwargs)

        if data is None:
            return

        selfmin = self.min
        if isinstance(selfmin, ThisSchema):
            selfmin = None

        if selfmin is not None and selfmin > data:
            raise ValueError(
                'Data {0} must be greater or equal than {1}'.format(
                    data, selfmin
                )
            )

        selfmax = self.max
        if isinstance(selfmax, ThisSchema):
            selfmax = None

        if selfmax is not None and selfmax < data:
            raise ValueError(
                'Data {0} must be lesser or equal than {1}'.format(
                    data, self.max
                )
            )


class LongSchema(NumberSchema):
    """Long Schema."""

    __data_types__ = [int] if PY3 else [long]
    default = 0 if PY3 else long(0)


class IntegerSchema(NumberSchema):
    """Integer Schema."""

    __data_types__ = [int]
    default = 0


class ComplexSchema(NumberSchema):
    """Complex Schema."""

    __data_types__ = [complex]
    default = 0j


class FloatSchema(NumberSchema):
    """Float Schema."""

    __data_types__ = [float]
    default = 0.


class StringSchema(ElementarySchema):
    """String Schema."""

    __data_types__ = [string_types]
    default = ''


class TypeSchema(ElementarySchema):
    """Type schema."""

    __data_types__ = [type]
    default = object


class ArraySchema(ElementarySchema):
    """Array Schema."""

    __data_types__ = [list, tuple, set]

    #: item types. Default any.
    itemtype = AnySchema()
    #: minimal array size. Default None.
    minsize = IntegerSchema(nullable=True, default=None)
    #: maximal array size. Default None.
    maxsize = IntegerSchema(nullable=True, default=None)
    unique = False  #: are items unique ? False by default.
    default = DynamicValue(lambda: [])

    def _validate(self, data, *args, **kwargs):

        ElementarySchema._validate(self, data, *args, **kwargs)

        if data is None:
            return

        selfminsize = self.minsize
        if isinstance(selfminsize, ThisSchema):
            selfminsize = None

        if selfminsize is not None and selfminsize > len(data):
            raise ValueError(
                'length of data {0} must be greater than {1}.'.format(
                    data, self.minsize
                )
            )

        selfmaxsize = self.maxsize
        if isinstance(selfmaxsize, ThisSchema):
            selfmaxsize = None

        if selfmaxsize is not None and len(data) > selfmaxsize:
            raise ValueError(
                'length of data {0} must be lesser than {1}.'.format(
                    data, self.maxsize
                )
            )

        if data:
            if isinstance(data, DynamicValue):
                data = data()

            if self.unique and len(set(data)) != len(data):
                raise ValueError('Duplicated items in {0}'.format(data))

            if self.itemtype is not None:
                for index, item in enumerate(data):
                    try:
                        validate(self.itemtype, item, owner=self.itemtype)

                    except Exception:
                        raise TypeError(
                            'Wrong type of {0} at {1}. {2} expected.'.format(
                                item, index, self.itemtype
                            )
                        )


class DictSchema(ArraySchema):
    """Array Schema."""

    __data_types__ = [dict]

    #: value type
    valuetype = AnySchema()
    default = DynamicValue(lambda: {})

    def _validate(self, data, *args, **kwargs):

        super(DictSchema, self)._validate(data, *args, **kwargs)

        if data:
            if self.unique and len(set(data.values())) != len(data):
                raise ValueError('Duplicated items in {0}'.format(data))

            if self.valuetype is not None:
                for key, item in data.items():
                    try:
                        validate(self.valuetype, item, owner=self.valuetype)

                    except Exception:
                        raise TypeError(
                            'Wrong type of {0} at {1}. {2} expected.'.format(
                                item, key, self.valuetype
                            )
                        )


class EnumSchema(ElementarySchema):
    """Enumerable schema."""

    __data_types__ = [Enum]


class DateTimeSchema(ElementarySchema):
    """Date time schema."""

    __data_types__ = [datetime]

    ms = IntegerSchema(nullable=True, default=None)
    s = IntegerSchema(nullable=True, default=None)
    mn = IntegerSchema(nullable=True, default=None)
    hr = IntegerSchema(nullable=True, default=None)
    day = IntegerSchema(nullable=True, default=None)
    month = IntegerSchema(nullable=True, default=None)
    year = IntegerSchema(nullable=True, default=None)

    default = DynamicValue(lambda: datetime.now())


class OneOfSchema(Schema):
    """Validate data with at least one schema."""

    schemas = ArraySchema(itemtype=AnySchema())  #: schemas to check

    def _validate(self, data, *args, **kwargs):

        super(OneOfSchema, self)._validate(data=data, *args, **kwargs)

        if self.nullable and data is None:
            return

        for selftype in self.schemas:
            try:
                validate(selftype, data)

            except Exception:
                continue

            else:
                break

        else:
            raise TypeError(
                '{0} does not match one of {1}'.format(data, self.schemas)
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Elementary UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..utils import validate

from ..elementary import (
    IntegerSchema, FloatSchema, ComplexSchema, LongSchema,
    StringSchema,
    ArraySchema,
    TypeSchema,
    DictSchema,
    OneOfSchema
)


class ElementaryTest(UTCase):

    __schemacls__ = None

    def _assert(self, data, error=False, **kwargs):

        if error:
            self.assertRaises(
                Exception, self.__schemacls__, default=data, **kwargs
            )

        else:
            schema = self.__schemacls__(default=data, **kwargs)
            validate(schema, data)

    def test_default(self):

        if self.__schemacls__ is None:
            return

        schema = self.__schemacls__()

        data = schema.default

        validate(schema, data=data)

    def test_none(self):

        if self.__schemacls__ is None:
            return

        schema = self.__schemacls__()

        if schema.nullable:
            validate(schema, data=None)

        else:
            self.assertRaises(TypeError, validate, schema, data=None)


class NumberSchemaTest(ElementaryTest):

    def test_min(self):

        if self.__schemacls__ is None:
            return

        self._assert(
            min=self.__schemacls__.__data_types__[0](0),
            data=self.__schemacls__.__data_types__[0](-1),
            error=True
        )

    def test_max(self):

        if self.__schemacls__ is None:
            return

        self._assert(
            max=self.__schemacls__.__data_types__[0](0),
            data=self.__schemacls__.__data_types__[0](2),
            error=True
        )


class IntegerSchemaTest(NumberSchemaTest):

    __schemacls__ = IntegerSchema


class ComplexSchemaTest(NumberSchemaTest):

    __schemacls__ = ComplexSchema


class LongSchemaTest(NumberSchemaTest):

    __schemacls__ = LongSchema


class FloatSchemaTest(NumberSchemaTest):

    __schemacls__ = FloatSchema


class StringSchemaTest(ElementaryTest):

    __schemacls__ = StringSchema


class TypeSchemaTest(ElementaryTest):

    __schemacls__ = TypeSchema

    def test_cls(self):

        self._assert(data=type)


class ArraySchemaTest(ElementaryTest):

    __schemacls__ = ArraySchema

    def test_maxsize(self):

        self._assert(maxsize=0, data=[])

    def test_maxsize_1(self):

        self._assert(maxsize=1, data=[1])

    def test_maxsize_1_0(self):

        self._assert(maxsize=1, data=[])

    def test_maxsize_error(self):

        self._assert(maxsize=1, data=[1, 0], error=True)

    def test_minsize(self):

        self._assert(minsize=0, data=[])

    def test_minsize_0(self):

        self._assert(minsize=0, data=[1])

    def test_minsize_1(self):

        self._assert(minsize=1, data=[1])

    def test_minsize_error(self):

        self._assert(minsize=1, data=[], error=True)

    def test_itemtype(self):

        self._assert(data=[1, 2], itemtype=IntegerSchema())

    def test_itemtype_error(self):

        self._assert(data=[1, 2.], itemtype=IntegerSchema(), error=True)

    def test_unique(self):

        self._assert(data=[1, 2], unique=True)

    def test_unique_error(self):

        self._assert(data=[1, 1], unique=True, error=True)


class DictSchemaTest(ElementaryTest):

    __schemacls__ = DictSchema

    def test_maxsize(self):

        self._assert(maxsize=0, data={})

    def test_maxsize_1(self):

        self._assert(maxsize=1, data={None: None})

    def test_maxsize_1_0(self):

        self._assert(maxsize=1, data={})

    def test_maxsize_error(self):

        self._assert(maxsize=1, data={None: None, 1: 1}, error=True)

    def test_minsize(self):

        self._assert(minsize=0, data={})

    def test_minsize_0(self):

        self._assert(minsize=0, data={None: None})

    def test_minsize_1(self):

        self._assert(minsize=1, data={None: None})

    def test_minsize_error(self):

        self._assert(minsize=1, data={}, error=True)

    def test_itemtype(self):

        self._assert(data={1: 1, 2: 2}, itemtype=IntegerSchema())

    def test_itemtype_error(self):

        self._assert(
            data={1: None, 2.: None}, itemtype=IntegerSchema(), error=True
        )

    def test_valuetype(self):

        self._assert(data={1: 1, 2.: 2}, valuetype=IntegerSchema())

    def test_valuetype_error(self):

        self._assert(data={1: 1, 2: 2.}, valuetype=IntegerSchema(), error=True)

    def test_unique(self):

        self._assert(data={1: 1, 2: 2}, unique=True)

    def test_unique_error(self):

        self._assert(data={1: 1, 2: 1}, unique=True, error=True)


class OneOfSchemaTest(UTCase):

    def test_default(self):

        schema = OneOfSchema()

        self.assertRaises(TypeError, validate, schema, 2)

    def test_error(self):

        schema = OneOfSchema(schemas=[IntegerSchema(), StringSchema()])

        self.assertRaises(TypeError, validate, schema, 3.)

    def test_success(self):

        schema = OneOfSchema(schemas=[IntegerSchema(), FloatSchema()])

        validate(schema, 3.)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Schema registry module."""

from six import iteritems

__all__ = [
    'SchemaRegistry',
    'getbyuuid', 'getbyname', 'register', 'unregister',
    'getschemasbyuuid', 'registercls', 'unregistercls', 'getbydatatype'
]


class SchemaRegistry(object):
    """In charge of register schemas."""

    def __init__(
            self,
            schemas=None, schbyname=None, schbyuuid=None, schbytype=None,
            *args, **kwargs
    ):
        """
        :param list schemas: starting schemas to register.
        :param dict schbyname: schemas by name.
        :param dict schyuid: schemas by uuid.
        :param dict schbytype: schemas by type.
        """
        super(SchemaRegistry, self).__init__(*args, **kwargs)

        self._schbyname = schbyname or {}
        self._schbyuuid = schbyuuid or {}
        self._schbytype = schbytype or {}

        if schemas:
            for schema in schemas:
                self.register(schema=schema)

    def getschemasbyuuid(self):
        """Get all schemas by uuid.

        :rtype: dict
        """
        return dict(self._schbyuuid)

    def register(self, schema):
        """Register input schema class.

        When registering a schema, all inner schemas are registered as well.

        :param Schema schema: schema to register.
        :return: old registered schema.
        :rtype: type
        """
        result = None

        uuid = schema.uuid

        if uuid in self._schbyuuid:
            result = self._schbyuuid[uuid]

        if result != schema:

            self._schbyuuid[uuid] = schema

            name = schema.name

            schemas = self._schbyname.setdefault(name, set())

            schemas.add(schema)

            for innername, innerschema in iteritems(schema.getschemas()):

                if innerschema.uuid not in self._schbyuuid:
                    register(innerschema)

        return result

    def registercls(self, data_types, schemacls=None):
        """Register schema class with associated data_types.

        Can be used such as a decorator.

        :param list data_types: data types to associate with schema class.
        :param type schemacls: schema class to register.
        :return: schemacls.
        :rtype: type
        """
        if schemacls is None:
            return lambda schemacls: self.registercls(
                data_types=data_types, schemacls=schemacls
            )

        for data_type in data_types:
            self._schbytype[data_type] = schemacls

        return schemacls

    def unregister(self, uuid):
        """Unregister a schema registered with input uuid.

        :raises: KeyError if uuid is not already registered.
        """
        schema = self._schbyuuid.pop(uuid)

        # clean schemas by name
        self._schbyname[schema.name].remove(schema)
        if not self._schbyname[schema.name]:
            del self._schbyname[schema.name]

    def unregistercls(self, schemacls=None, data_types=None):
        """Unregister schema class or associated data_types.

        :param type schemacls: sub class of Schema.
        :param list data_types: data_types to unregister.
        """
        if schemacls is not None:

            # clean schemas by data type
            for data_type in list(self._schbytype):
                _schemacls = self._schbytype[data_type]
                if _schemacls is schemacls:
                    del self._schbytype[data_type]

        if data_types is not None:

            for data_type in data_types:
                if data_type in self._schbytype:
                    del self._schbytype[data_type]

    def getbyuuid(self, uuid):
        """Get a schema by given uuid.

        :param str uuid: schema uuid to retrieve.
        :rtype: Schema
        :raises: KeyError if uuid is not registered already.
        """
        if uuid not in self._schbyuuid:
            raise KeyError('uuid {0} not registered'.format(uuid))

        return self._schbyuuid[uuid]

    def getbyname(self, name):
        """Get schemas by given name.

        :param str name: schema names to retrieve.
        :rtype: list
        :raises: KeyError if name is not registered already.
        """
        if name not in self._schbyname:
            raise KeyError('name {0} not registered'.format(name))

        return self._schbyname[name]

    def getbydatatype(self, data_type, besteffort=True):
        """Get schema class by data type.

        :param type data_type: data type from where get schema class.
        :param bool besteffort: if True and data_type not registered, parse all
            registered data_types and stop when once data_type is a subclass of
            input data_type.
        :return: sub class of Schema.
        :rtype: type
        """
        result = None

        if data_type in self._schbytype:
            result = self._schbytype[data_type]

        elif besteffort:
            for rdata_type in self._schbytype:
                if issubclass(data_type, rdata_type):
                    result = self._schbytype[rdata_type]
                    break

        return result


_REGISTRY = SchemaRegistry()  #: global Schemaregistry.


def register(schema):
    """Register globally input shema.

    :param b3j0f.schema.Schema: schema to register.
    """
    return _REGISTRY.register(schema=schema)


def unregister(uuid):
    """Unregister."""
    return _REGISTRY.unregister(uuid=uuid)


def getbyuuid(uuid):
    """Get globally a schema by given uuid.

    :param str uuid: schema uuid to retrieve.
    :rtype: type
    :raises: KeyError if name is not registered already.
    """
    return _REGISTRY.getbyuuid(uuid=uuid)


def getbyname(name):
    """Get globally a schema by given name.

    :param str name: schema names to retrieve.
    :rtype: list
    :raises: KeyError if name is not registered already.
    """
    return _REGISTRY.getbyname(name=name)


def getschemasbyuuid():
        """Get all schemas by uuid.

        :rtype: dict
        """
        return _REGISTRY.getschemasbyuuid()


def getbydatatype(data_type, besteffort=True):
    """Get schema class by data type.

    :param type data_type: data type from where get schema class.
    :param bool besteffort: if True and data_type not registered, parse all
        registered data_types and stop when once data_type is a subclass of
        input data_type.
    :return: sub class of Schema.
    :rtype: type
    """
    return _REGISTRY.getbydatatype(data_type=data_type, besteffort=besteffort)


def registercls(data_types, schemacls=None):
    """Register schema class with associated data_types.

    Can be used such as a decorator.

    :param list data_types: data types to associate with schema class.
    :param type schemacls: schema class to register.
    :return: schemacls.
    :rtype: type
    """
    return _REGISTRY.registercls(data_types=data_types, schemacls=schemacls)


def unregistercls(self, schemacls=None, data_types=None):
    """Unregister schema class or associated data_types.

    :param type schemacls: sub class of Schema.
    :param list data_types: data_types to unregister.
    """
    return _REGISTRY.unregistercls(schemacls=schemacls, data_types=data_types)
